Method: Timecode#with_srt_fraction

Defined in:
lib/timecode.rb

#with_srt_fractionObject

SRT uses a fraction of a second as the last element instead of number of frames, with a comma as the separator

Timecode.parse("00:00:10:24", 25).with_srt_fraction #=> "00:00:10,96"


473
474
475
# File 'lib/timecode.rb', line 473

def with_srt_fraction
  with_frames_as_fraction(WITH_SRT_FRACTION)
end